/*
 *      run.cpp
 *
 *      Module to run the ME 161 Haptic Display
 *      Users can change the feedback values
 *      and specify step input and data saving
 *
 *      -Allison Okamura 11/18/96
 *      -Allison Okamura 11/17/98
 */

/* Includes */
#include <stdio.h>
#include <dos.h>
#include <conio.h>
#include <iostream.h>
#include <math.h>
#include "haptic.h"
#include "ioboard.h"
#include "timing.h"
#include "readcfg.h"

/* Defines */
#define NUM       4
#define GROUPNUM  0
#define ENABLED   1
#define KP        2
#define KD        3
#define STEP_VAL  4
#define PARAM_NUM 5
#define OFF       0
#define ON        1
#define LO        0
#define HI        1
#define MINUS     0
#define PLUS      1
#define DATA_SEC  3    // Length of time (sec) to take data
#define FREQ      1000 // Hz
#define USER_TIMER_TICK                 0x1C // interrupt for timing
#define MAXDATA   1000
#define WRITE     10   // number of cycles before writing to motors

/* Globals */
void interrupt ( *oldhandler)(...);
void interrupt RunISR(...);

int Station = 0;
int Param = GROUPNUM;
int ApplyStep = OFF;
int TakeData = OFF;

int Grp_num[NUM];
int Enabled[NUM] = {OFF, OFF, OFF, OFF};
float Kp[NUM] = {0.0, 0.0, 0.0, 0.0};
float Kd[NUM] = {0.0, 0.0, 0.0, 0.0};
float StepValue[NUM] = {0.0, 0.0, 0.0, 0.0};
float a[NUM] = {0.0, 0.0, 0.0, 0.0};
float b[NUM] = {0.0, 0.0, 0.0, 0.0};
float c[NUM] = {0.0, 0.0, 0.0, 0.0};
float d[NUM] = {0.0, 0.0, 0.0, 0.0};
int Calibrated[NUM] = {OFF, OFF, OFF, OFF};

float freq;
float x[NUM], dx[NUM], output[NUM];
int time_ctr = 0;
int data_ctr = 1;
int calc_ctr = 0;
float x_data[MAXDATA];
float step_data[MAXDATA];
float time_data[MAXDATA];

/*------------------------- RunISR -------------------------*
 *  Interrupt Service Routine for running haptic displays   *
 *----------------------------------------------------------*/
void interrupt RunISR(...)
{
	 int i, j;
	 float sensor[NUM][WRITE], avg_sensor[NUM];
	 float angle, current[NUM];
    static float last_x[NUM] = {0.0, 0.0};

	 // Read Sensors
	 for (i=0; i<NUM; i++) sensor[i][calc_ctr] = (float) read_ad();

	 calc_ctr++;

	 if (calc_ctr >= WRITE) {

		 calc_ctr = 0;

		 // for each station
		 for (i=0; i<NUM; i++) {

		  // if step response had been applied, add it
			if (i != Station) current[i] = 0;
			else current[i] = ApplyStep*StepValue[i];

			// get the average sensor readings
			avg_sensor[i] = 0;
			for (j=0; j<WRITE; j++) avg_sensor[i] += sensor[i][j];
			avg_sensor[i] *= 1.0/WRITE;

			angle = a[i]*avg_sensor[i]*avg_sensor[i]*avg_sensor[i]
					+ b[i]*avg_sensor[i]*avg_sensor[i]
					+ c[i]*avg_sensor[i] + d[i] - 180;

			// get position and velocity
			x[i] = angle*DEG2RAD*LX;
			dx[i] = (x[i] - last_x[i])*freq;
			last_x[i] = x[i];

			current[i] += -Kp[i]*x[i] - Kd[i]*dx[i];  // feedback current
			output[i] = current[i]*Enabled[i];        // calculate output volts
			output_da(i,output[i]);                   // send to motor
		 }

		 // Data Storage
		 if (TakeData) {
			time_ctr++;
			if (time_ctr > (DATA_SEC*freq/WRITE)) TakeData = OFF;
			data_ctr++;
			if (data_ctr < MAXDATA) {
				time_data[data_ctr] = ((float) time_ctr)/freq*WRITE;
				x_data[data_ctr] = x[Station];
				step_data[data_ctr] = ApplyStep*StepValue[Station];
			}
		}
	 }
}

/*------------------------ Outline -------------------------*
 *  Draws outline for running haptic displays               *
 *----------------------------------------------------------*/
void Outline(void)
{
	 clrscr();

	 // Instructions
	 textcolor(WHITE);
	 gotoxy(1,1); cprintf("ME 161 Haptic Paddle - Feedback Control");
	 textcolor(CYAN);
	 gotoxy(1,3); cprintf("Use arrows on keypad to select station number and parameter.");
	 gotoxy(1,4); cprintf("Use + and - to change parameters, or hit one of the following keys:");
	 gotoxy(1,5); cprintf("(a)pply step input               (r)emove step input");
	 gotoxy(1,6); cprintf("(t)ake data for 3 seconds        (q)uit");

	 // Labels
	 textcolor(CYAN);
	 gotoxy( 1,8); cprintf("Sta");
	 gotoxy( 5,8); cprintf("Grp");
	 gotoxy(10,8); cprintf("Enabled");
	 gotoxy(20,8); cprintf("C2 (A/m)");
	 gotoxy(30,8); cprintf("C1 (As/m)");
	 gotoxy(42,8); cprintf("Step (mA)");
	 gotoxy(55,8); cprintf("x (mm)");
	 gotoxy(65,8); cprintf("Output (mA)");

	 // Station Numbers
	 textcolor(CYAN);
	 gotoxy( 1,10); cprintf("1");
	 gotoxy( 1,12); cprintf("2");
	 gotoxy( 1,14); cprintf("3");
	 gotoxy( 1,16); cprintf("4");

	 gotoxy(1,22); cprintf("Status:");
}

/*------------------- DisplayParams ------------------------*
 *  Show Parameters for running haptic displays             *
 *----------------------------------------------------------*/
void DisplayParams(void)
{
	 int i;
	 // first reprint the whole display

	 textcolor(WHITE);
	 for (i=0; i<NUM; i++) {
		 // Group number
		 gotoxy( 5,10+2*i); if(Calibrated[i]) cprintf(" %d  ",Grp_num[i]); else cprintf("None");

		 // Enabled
		 gotoxy(10,10+2*i); if (Enabled[i]) cprintf("  On "); else cprintf("  Off ");

		 // Kp
		 gotoxy(20,10+2*i); cprintf("%3.1f ",Kp[i]);

		 // Kd
		 gotoxy(30,10+2*i); cprintf("%1.3f ",Kd[i]);

		 // Step Value
		 gotoxy(42,10+2*i); cprintf("%3.1f   ",StepValue[i]);

	 }

	 textcolor(GREEN);

	 // now highlight the new parameter
	 switch(Param) {
		case GROUPNUM:
			gotoxy( 5,10+2*Station); if(Calibrated[Station]) cprintf(" %d  ",Grp_num[Station]); else cprintf("None");
			break;
		case ENABLED:
			gotoxy(10,10+2*Station); if (Enabled[Station]) cprintf("  On "); else cprintf("  Off ");
			break;
		case KP:
			gotoxy(20,10+2*Station); cprintf("%3.1f ",Kp[Station]);
			break;
		case KD:
			gotoxy(30,10+2*Station); cprintf("%1.3f ",Kd[Station]);
			break;
		case STEP_VAL:
			gotoxy(42,10+2*Station); cprintf("%3.1f    ",StepValue[Station]);
			break;
		default:
			break;
	}
}

/*--------------------- EnterGrpNum ------------------------*
 *  Reads calibration values                                *
 *----------------------------------------------------------*/
void EnterGrpNum(void)
{
	char string[40];
	char astring[4],  bstring[4], cstring[4], dstring[4];

	textcolor(CYAN);
	gotoxy(1,22); cprintf("Enter in your group number: ");
	gets(string);
	sscanf(string, "%i", &Grp_num[Station]);

	// Read calibration values from data file
	sprintf(astring, "A%i", Grp_num[Station]);
	sprintf(bstring, "B%i", Grp_num[Station]);
	sprintf(cstring, "C%i", Grp_num[Station]);
	sprintf(dstring, "D%i", Grp_num[Station]);
	ReadCfgAddFloat(astring,&a[Station],0);
	ReadCfgAddFloat(bstring,&b[Station],0);
	ReadCfgAddFloat(cstring,&c[Station],0);
	ReadCfgAddFloat(dstring,&d[Station],0);
	ReadCfgExecute("calib.dat",RCE_APPLY_DEFAULTS);
	ReadCfgClear();                         // clear config variable list

	if ((a[Station] != 0) && (b[Station] != 0) && (c[Station] != 0) && (d[Station] != 0)) {
		gotoxy(1,22); cprintf("Status:"); clreol();
		gotoxy(1,23); cprintf("Calibration values entered"); clreol();
		Calibrated[Station] = 1;
	} else {
		gotoxy(1,22); cprintf("Status:"); clreol();
		gotoxy(1,23); cprintf("Bad calibration values. Enter a new group number."); clreol();
		Calibrated[Station] = 0;
	}

	textcolor(GREEN);
}

/*------------------- ChangeParam  -------------------------*
 *  Changes Parameter for running haptic displays           *
 *----------------------------------------------------------*/
void ChangeParam(int sign)
{
	textcolor(GREEN);

	 // change the select parameter
	switch(Param) {
	case GROUPNUM:
		if (Enabled[Station] == ON) {
			textcolor(RED); gotoxy(1,23);
			cprintf("You must disable your station first"); clreol();
		} else EnterGrpNum();
		gotoxy( 5,10+2*Station); if(Calibrated[Station]) cprintf(" %d  ",Grp_num[Station]); else cprintf("None");
		break;
	case ENABLED:
		if (sign) {
			if (Calibrated[Station]) Enabled[Station] = ON;
			else {
				Enabled[Station] = OFF;
				textcolor(RED); gotoxy(1,23);
				cprintf("You must enter a group number first."); clreol();
			}
		} else Enabled[Station] = OFF;
		gotoxy(10,10+2*Station); if (Enabled[Station]) cprintf("  On "); else cprintf("  Off ");
		break;
	case KP:
		if (sign) Kp[Station] += 1.0;
		else Kp[Station] -= 1.0;
		gotoxy(20,10+2*Station); cprintf("%3.1f ",Kp[Station]);
		break;
	case KD:
		if (sign) Kd[Station] += 0.005;
		else Kd[Station] -= 0.005;
		gotoxy(30,10+2*Station); cprintf("%1.3f ",Kd[Station]);
		break;
	case STEP_VAL:
		if (sign) StepValue[Station] += 10;
		else StepValue[Station] -= 10;
		gotoxy(42,10+2*Station); cprintf("%3.1f    ",StepValue[Station]);
		break;
	default :
		break;
	 }
}

/*-------------------------- Run ---------------------------*
 *  Main Loop for running                                   *
 *----------------------------------------------------------*/
void Run(void)
{
	 char key;
	 int done = OFF;
	 int temp, i;
	 static int LastTakeData = OFF;
	 float position[NUM], iout[NUM];
	 FILE *fp;
	 char filename[10], mfilename[15];

	 // initialize display
	 init_board();
	 clrscr();
	 Outline();
	 DisplayParams();

	 // set the number of channels to scan
	 scan_channels(0,(NUM-1));
	 for (i=0; i<NUM; i++) Enabled[i] = OFF;

	 // start interrupt
	 freq = setTimer(FREQ);
	 oldhandler = getvect(USER_TIMER_TICK);
	 setvect(USER_TIMER_TICK, RunISR);
	 gotoxy(1,11); clreol();
	 enable();

	// continuous loop
	while (!done) {

		// make local copy of interrupt variable
		disable();
		for (i=0; i<NUM; i++) {
			position[i] = x[i];
			iout[i] = output[i];
		}
		enable();

		for (i=0; i<NUM; i++) {
			if (Enabled[i]) {
				gotoxy(55,10+2*i);
				printf("%3.2f   ", position[i]);
				gotoxy(65,10+2*i);
				printf("%3.2f   ", iout[i]);
			 } else {
				gotoxy(55,10+2*i);
				printf("          ");
				gotoxy(65,10+2*i);
				printf("          ");
			 }
		}

		if (!TakeData && LastTakeData) {
			// Open file
			gotoxy(1,24); cprintf("Enter a filename (8 char max, do not include .m extension): ");
			gets(filename);
			sprintf(mfilename, "%s.m", filename);
			fp = fopen(mfilename, "w");
			if (fp == NULL) {
				gotoxy(1,23); textcolor(RED);
				cprintf("Error opening file."); clreol();
			} else {
				for (i=1; i<data_ctr; i++) {
					fprintf(fp, "time(%d) = %f; \t", i, time_data[i]);
					fprintf(fp, "x(%d) = %f; \t", i, x_data[i]);
					fprintf(fp, "step(%d) = %f;\n", i, step_data[i]);
				}
				data_ctr = 1;
				gotoxy(1,23); textcolor(RED);
				cprintf("Done taking Data. Saved in file %s.", mfilename); clreol();
				gotoxy(1,24); clreol();
			}
		}
		LastTakeData = TakeData;

		if (kbhit()) {
			key = getch();
			switch (key) {
			case '2':
				 temp = Station + 1;
				 if (temp > NUM-1) Station = NUM-1;
				 else Station = temp;
				 DisplayParams();
				 break;
			case '8':
				 temp = Station - 1;
				 if (temp < 0) Station = 0;
				 else Station = temp;
				 DisplayParams();
				 break;
			case '4':
				 temp = Param - 1;
				 if (temp < 0) Param = 0;
				 else Param = temp;
				 DisplayParams();
				 break;
			case '6':
				 temp = Param + 1;
				 if (temp > PARAM_NUM-1) Param = PARAM_NUM-1;
				 else Param = temp;
				 DisplayParams();
				 break;
			case '+':
				 ChangeParam(PLUS);
				 break;
			case '-':
				 ChangeParam(MINUS);
				 break;
			case 'a':
				 ApplyStep = ON;
				 gotoxy(1,23);
				 textcolor(RED);
				 cprintf("Step input applied."); clreol();
				 break;
			case 'r':
				 ApplyStep = OFF;
				 gotoxy(1,23);
				 textcolor(RED);
				 cprintf("Step input removed."); clreol();
				 break;
			case 't':
				 disable();
				 TakeData = ON;
				 time_ctr = 0;
				 enable();
				 gotoxy(1,23);
				 textcolor(RED);
				 cprintf("Taking Data."); clreol();
				 break;
			case 'q':
				done = ON;
				break;
			default:
				break;
			}
		}
	}

	// end interrupt
	disable();
	setvect(USER_TIMER_TICK,oldhandler);

	 // zero output
	 for (i=0; i<NUM; i++) output_da(i,0.0);
}
