/*
 *      tether.cpp
 *
 *      Module to run the ME 161 Haptic Display
 *                Tetherball game.
 *
 *      -Christopher Richard and Allison Okamura 12/1/96
 *      -Allison Okamura and Weston Griffin 11/25/98
 *
 */

/* Includes */
#include <stdio.h>
#include <dos.h>
#include <conio.h>
#include <iostream.h>
#include <math.h>
#include <graphics.h>
#include "readcfg.h"
#include "display.h"
#include "haptic.h"
#include "ioboard.h"
#include "timing.h"
#include "dynamics.h"

/* Defines */
#define NUM       4
#define OFF       0
#define ON        1
#define LO        0
#define HI        1
#define FREQ      5000 // Hz
#define TIMER           0x1C // interrupt for timing
#define WRITE     10   // number of cycles before writing to motors
#define RADIUS    75
#define KD        0.001

/* Globals */
void interrupt (*Toldhandler)(...);
void interrupt TetherBallISR(...);

int TCalibrated[NUM] = {OFF, OFF, OFF, OFF};
float Ta[NUM] = {0.0, 0.0, 0.0, 0.0};
float Tb[NUM] = {0.0, 0.0, 0.0, 0.0};
float Tc[NUM] = {0.0, 0.0, 0.0, 0.0};
float Td[NUM] = {0.0, 0.0, 0.0, 0.0};
int TStation = 0;
int LDEnabled = OFF;
int RDEnabled = OFF;

float Tfreq;
float theta[NUM], Toutput[NUM];
extern float fxp, fxn, fyp, fyn;
extern float thetaX, dthetaX, ddthetaX, dthetaXold, ddthetaXold;
extern float thetaY, dthetaY, ddthetaY, dthetaYold, ddthetaYold;

/*-------------------- TetherBallISR -----------------------*
 *  Interrupt Service Routine for tetherball game           *
 *----------------------------------------------------------*/
void interrupt TetherBallISR(...)
{
	 int i, j;
    float sensor[NUM][WRITE], avg_sensor[NUM];
	 float angle, current[NUM];
    static float last_theta[NUM] = {0.0, 0.0, 0.0, 0.0};
	 static int calc_ctr = 0;
	 float dtheta[NUM];

    // Read Sensors
    for (i=0; i<NUM; i++) sensor[i][calc_ctr] = (float) read_ad();

	 calc_ctr++;

    if (calc_ctr >= WRITE) {

		 calc_ctr = 0;

		 // for each station
		 for (i=0; i<NUM; i++) {

			// get the average sensor readings
	 avg_sensor[i] = 0;
	 for (j=0; j<WRITE; j++) avg_sensor[i] += sensor[i][j];
			avg_sensor[i] *= 1.0/WRITE;

		angle = Ta[i]*avg_sensor[i]*avg_sensor[i]*avg_sensor[i]
			 + Tb[i]*avg_sensor[i]*avg_sensor[i]
			 + Tc[i]*avg_sensor[i] + Td[i] - 180.0;

		if (TCalibrated[i] == OFF) angle = 0;

	// get position and velocity
	 theta[i] = angle;
	 dtheta[i] = (theta[i] - last_theta[i])*Tfreq*WRITE;
			last_theta[i] = theta[i];
		 }

		// Calculate dynamic interaction
		tetherDyn(WRITE/Tfreq, LDEnabled, RDEnabled);
		current[0] = fxn;
		current[1] = fxp;
		current[2] = fyn;
		current[3] = fyp;

		for (i=0; i<2; i++) {
			current[i] += -KD*dtheta[i];
			if (LDEnabled) {
				if (i == 0) Toutput[i] = current[i]; // calculate output current
				else Toutput[i] = -current[i];
			}
			else Toutput[i] = 0;
			output_da(i,Toutput[i]);             // send to motor
		}
		for (i=2; i<NUM; i++) {
			current[i] += -KD*dtheta[i];
			if (RDEnabled) {
         	if (i == 2) Toutput[i] = current[i]; // calculate output current
            else Toutput[i] = -current[i];
         }
			else Toutput[i] = 0;
			output_da(i,Toutput[i]);             // send to motor
		}
	}
}

/*--------------------- ReadCal ----------------------------*
 *  Reads calibration values                                *
 *----------------------------------------------------------*/
void ReadCal(int station)
{
	char string1[20], string2[60];
	char astring[3], bstring[3], cstring[3], dstring[3];
	int grp_num;
	int polypoints[8] = {1,120,400,120,400,180,1,180};

	setcolor(CYAN);
	sprintf(string1, "Enter the group number for station %d: ", station + 1);
	outtextxy(1,130, string1);
	setcolor(WHITE);
	gotoxy(1,10); gets(string2);
	sscanf(string2, "%i", &grp_num);
	setcolor(BLACK);
	setlinestyle(SOLID_LINE,0,NORM_WIDTH);
	setfillstyle(SOLID_FILL, BLACK);
	fillpoly(4, polypoints);

	// Read calibration values from data file
	sprintf(astring, "A%i", grp_num);
	sprintf(bstring, "B%i", grp_num);
	sprintf(cstring, "C%i", grp_num);
	sprintf(dstring, "D%i", grp_num);
	ReadCfgAddFloat(astring,&Ta[station],0);
	ReadCfgAddFloat(bstring,&Tb[station],0);
	ReadCfgAddFloat(cstring,&Tc[station],0);
	ReadCfgAddFloat(dstring,&Td[station],0);
	ReadCfgExecute("calib.dat",RCE_APPLY_DEFAULTS);
	ReadCfgClear();                         // clear config variable list

	//if ((a[Station] != 0) && (b[Station] != 0) && (c[Station] != 0) && (d[Station] != 0)) {
	//} else {
	//}
}

/*--------------------- TetherOutline ----------------------*
 *  Draws outline for running the tetherball game           *
 *----------------------------------------------------------*/
void TetherOutline(void)
{
    setcolor(WHITE);
	 outtextxy(10,10,"ME 161 Haptic TetherBall Game");

    // Instructions
    setcolor(CYAN);
    outtextxy(10,30,"Use 1-4 to select station:");
    outtextxy(10,40,"(c)alibration");

	 outtextxy(10,60,"(l)eft dynamics enable/disable");
	 outtextxy(10,70,"(r)ight dynamics enable/disable");
	 outtextxy(10,90,"(s)tart over");
	 outtextxy(10,100,"(q)uit");

	 // Current Settings
    setcolor(GREEN);
	 outtextxy(270,30,"1");
	 outtextxy(270,40,"calibrated");
	 outtextxy(270,60,"disabled");
	 outtextxy(270,70,"disabled");

	 // Current position
	 setcolor(YELLOW);
	 outtextxy(400,30,"position:");
	 outtextxy(400,40,"output:");
	 outtextxy(400,50,"ball:");

    // Draw tetherball 1
	 setcolor(YELLOW);
    setlinestyle(SOLID_LINE,0,NORM_WIDTH);
    rectangle(130,350,190,410);   // base
    setfillstyle(SOLID_FILL, WHITE);
    fillellipse(160,350-RADIUS,10,10); // ball
    setlinestyle(SOLID_LINE,0,THICK_WIDTH);
	 line(160,350,160,350-RADIUS);      // radius
    setcolor(GREEN);
	 line(65,350,105,350); // paddles
	 line(215,350,255,350);

	 // Draw tetherball 2
	 setcolor(YELLOW);
	 setlinestyle(SOLID_LINE,0,NORM_WIDTH);
	 rectangle(450,350,510,410);   // base
	 setfillstyle(SOLID_FILL, WHITE);
	 fillellipse(480,350-RADIUS,10,10); // ball
	 setlinestyle(SOLID_LINE,0,THICK_WIDTH);
	 line(480,350,480,350-RADIUS);      // radius
	 setcolor(GREEN);
	 line(385,350,425,350); // paddles
	 line(535,350,575,350);
}

/*--------------------- DisplayGraphics --------------------*
 *  Draws graphics for tetherball game                      *
 *----------------------------------------------------------*/
void DisplayGraphics(void)
{
	int i;
   char string[20];
   static float ball_pos[2], last_ball_print = 0.0;
	static float pos[NUM], last_pos[NUM] = {0.0, 0.0, 0.0, 0.0}, last_pos_print = 0.0;
   static float iout[NUM], last_iout = 0.0;
	static int ballx[2], bally[2], last_ballx[2] = {160,480}, last_bally[2] = {350-RADIUS,350-RADIUS};

	// make local copy of interrupt variables
	disable();
	ball_pos[0] = thetaX;
   ball_pos[1] = thetaY;
   for (i=0; i<NUM; i++) {
		pos[i] = theta[i];
      iout[i] = Toutput[i];
   }
	enable();

   // erase old data
	setcolor(BLACK);
	sprintf(string,"%3.1f",last_pos_print);
	outtextxy(500,30,string);
	sprintf(string,"%3.1f",last_iout);
	outtextxy(500,40,string);
	sprintf(string,"%3.1f",last_ball_print);
	outtextxy(500,50,string);

	// draw new data
	setcolor(YELLOW);
	sprintf(string,"%3.1f",pos[TStation]);
	outtextxy(500,30,string);
	sprintf(string,"%3.1f",iout[TStation]);
	outtextxy(500,40,string);
	if (TStation < 2) sprintf(string,"%3.1f",ball_pos[0]);
	else sprintf(string,"%3.1f",ball_pos[1]);
	outtextxy(500,50,string);

	// erase old tetherball 1 graphics
	setcolor(BLACK);
	setlinestyle(SOLID_LINE,0,THICK_WIDTH);
	line(160,350,last_ballx[0],last_bally[0]);              // radius
	line(65,350-(int)last_pos[0],155,350-(int)last_pos[0]); // paddles
	line(165,350-(int)last_pos[1],255,350-(int)last_pos[1]);
	setlinestyle(SOLID_LINE,0,NORM_WIDTH);
	setfillstyle(SOLID_FILL, BLACK);
	fillellipse(last_ballx[0],last_bally[0],10,10);         // ball

	// erase old tetherball 2 graphics
	setlinestyle(SOLID_LINE,0,THICK_WIDTH);
	line(480,350,last_ballx[1],last_bally[1]);              // radius
	line(385,350-(int)last_pos[2],475,350-(int)last_pos[2]); // paddles
	line(485,350-(int)last_pos[3],575,350-(int)last_pos[3]);
	setlinestyle(SOLID_LINE,0,NORM_WIDTH);
	fillellipse(last_ballx[1],last_bally[1],10,10);         // ball

	// draw new tetherball 1 graphics
	setcolor(YELLOW);
	setlinestyle(SOLID_LINE,0,NORM_WIDTH);
	rectangle(130,350,190,410);                         // base
	ballx[0] = (int) (160+RADIUS*sin(ball_pos[0]*DEG2RAD));
	bally[0] = (int) (350-RADIUS*cos(ball_pos[0]*DEG2RAD));
	setlinestyle(SOLID_LINE,0,THICK_WIDTH);
	line(160,350,ballx[0],bally[0]);                    // radius
	setcolor(GREEN);
	line(65,350-(int)pos[0],105,350-(int)pos[0]);       // paddles
	line(215,350-(int)pos[1],255,350-(int)pos[1]);
	setcolor(YELLOW);
	setlinestyle(SOLID_LINE,0,NORM_WIDTH);
	setfillstyle(SOLID_FILL, WHITE);
	fillellipse(ballx[0],bally[0],10,10);               // ball

	// draw new tetherball 2 graphics
	setcolor(YELLOW);
	setlinestyle(SOLID_LINE,0,NORM_WIDTH);
	rectangle(450,350,510,410);                    // base
	ballx[1] = (int) (480+RADIUS*sin(ball_pos[1]*DEG2RAD));
	bally[1] = (int) (350-RADIUS*cos(ball_pos[1]*DEG2RAD));
	setlinestyle(SOLID_LINE,0,THICK_WIDTH);
	line(480,350,ballx[1],bally[1]);               // radius
	setcolor(GREEN);
	line(385,350-(int)pos[2],425,350-(int)pos[2]); // paddles
	line(535,350-(int)pos[3],575,350-(int)pos[3]);
	setcolor(YELLOW);
	setlinestyle(SOLID_LINE,0,NORM_WIDTH);
	setfillstyle(SOLID_FILL, WHITE);
	fillellipse(ballx[1],bally[1],10,10);          // ball
	// store last positions
	last_ballx[0] = ballx[0];
	last_bally[0] = bally[0];
	last_ballx[1] = ballx[1];
	last_bally[1] = bally[1];
	if (TStation < 2)    last_ball_print = ball_pos[0];
	else last_ball_print = ball_pos[1];
	for (i=0; i<NUM; i++) {
		last_pos[i] = pos[i];
	}
	last_pos_print = pos[TStation];
	last_iout = iout[TStation];
}

/*------------------------ TetherBall ----------------------*
 *  Main Loop for tetherball game                           *
 *----------------------------------------------------------*/
void TetherBall(void)
{
    char key;
    int done = OFF;
    int i;

    thetaX = THETAX0;
	 thetaY = THETAY0;

    // set the number of channels to scan
	 TStation = 0;
	 scan_channels(0,(NUM-1));
	 LDEnabled = OFF;
	 RDEnabled = OFF;

    // read dynamics info
    initDyn();

    // draw initial screen
	 init_graphics();
	 TetherOutline();

	 // zero output
	 for (i=0; i<NUM; i++) output_da(i,0.0);

    // start interrupt
	 Tfreq = setTimer(FREQ);
	 Toldhandler = getvect(TIMER);
	 setvect(TIMER, TetherBallISR);
	 enable();

	// continuous loop
	while (!done) {

		// update graphics and printouts
		DisplayGraphics();

		// user input
		if (kbhit()) {
			key = getch();
			switch (key) {
				case 'c':
					if (TStation <2) LDEnabled = OFF;
					else RDEnabled = OFF;
					ReadCal(TStation);
					TCalibrated[TStation] = ON;
					setcolor(BLACK); outtextxy(270,40,"not calibrated");
					setcolor(GREEN); outtextxy(270,40,"calibrated");
					break;
				case 'l':
					if (LDEnabled) {
						LDEnabled = OFF;
						setcolor(BLACK); outtextxy(270,60,"enabled");
						setcolor(GREEN); outtextxy(270,60,"disabled");
					}
					else if ((TCalibrated[0] == ON) && (TCalibrated[1] == ON)) {
						LDEnabled = ON;
						setcolor(BLACK); outtextxy(270,60,"disabled");
						setcolor(GREEN); outtextxy(270,60,"enabled");
					}
					break;
				case 'r':
					if (RDEnabled) {
						RDEnabled = OFF;
						setcolor(BLACK); outtextxy(270,70,"enabled");
						setcolor(GREEN); outtextxy(270,70,"disabled");
					}
					else if ((TCalibrated[2] == ON) && (TCalibrated[3] == ON)) {
						RDEnabled = ON;
						setcolor(BLACK); outtextxy(270,70,"disabled");
						setcolor(GREEN); outtextxy(270,70,"enabled");
					}
					break;
				case '1':
					// erase old values
					setcolor(BLACK);
					outtextxy(270,30,"2");
					outtextxy(270,30,"3");
					outtextxy(270,30,"4");
					outtextxy(270,40,"calibrated");
					outtextxy(270,40,"not calibrated");
					// new values
					TStation = 0;
					setcolor(GREEN);
					outtextxy(270,30,"1");
					if (TCalibrated[TStation]) outtextxy(270,40,"calibrated");
					else outtextxy(270,40,"not calibrated");
					break;
				case '2':
					// erase old values
					setcolor(BLACK);
					outtextxy(270,30,"1");
					outtextxy(270,30,"3");
					outtextxy(270,30,"4");
					outtextxy(270,40,"calibrated");
					outtextxy(270,40,"not calibrated");
					// new values
					TStation = 1;
					setcolor(GREEN);
					outtextxy(270,30,"2");
					if (TCalibrated[TStation]) outtextxy(270,40,"calibrated");
					else outtextxy(270,40,"not calibrated");
					break;
				case '3':
					// erase old values
					setcolor(BLACK);
					outtextxy(270,30,"1");
					outtextxy(270,30,"2");
					outtextxy(270,30,"4");
					outtextxy(270,40,"calibrated");
					outtextxy(270,40,"not calibrated");
					// new values
					TStation = 2;
					setcolor(GREEN);
					outtextxy(270,30,"3");
					if (TCalibrated[TStation]) outtextxy(270,40,"calibrated");
					else outtextxy(270,40,"not calibrated");
					break;
				case '4':
					// erase old values
					setcolor(BLACK);
					outtextxy(270,30,"1");
					outtextxy(270,30,"2");
					outtextxy(270,30,"3");
					outtextxy(270,40,"calibrated");
					outtextxy(270,40,"not calibrated");
					// new values
					TStation = 3;
					setcolor(GREEN);
					outtextxy(270,30,"4");
					if (TCalibrated[TStation]) outtextxy(270,40,"calibrated");
					else outtextxy(270,40,"not calibrated");
					break;
				case 's':
					disable();
					thetaX = -10.0;
					dthetaX = 0.0;
					ddthetaX = 0.0;
					dthetaXold = 0.0;
					ddthetaXold = 0.0;
					thetaY = -10.0;
					dthetaY = 0.0;
					ddthetaY = 0.0;
					dthetaYold = 0.0;
					ddthetaYold = 0.0;
					enable();
					break;
				case 'q':
					done = ON;
					break;
				default:
					break;
			}
		}
	}

    // shut down
    getch();
	 disable();
    setvect(TIMER,Toldhandler);
    for (i=0; i<NUM; i++) output_da(i,0.0);
	 closegraph();
}
